# -*- coding: utf-8 -*-
import os
import glob
import logging
import threading
import traceback
from multiprocessing import Manager

logger = logging.getLogger(__name__)


class FileQueue:
    def __init__(self, path, max_head_index=100):
        self._path = path
        self._max_head_index = max_head_index
        self._manager = Manager()
        self._items = self._manager.list()
        self._mutex = threading.Lock()
        self._mutex.acquire()
        try:
            logger.info("initializing FileQueue at {}".format(self._path))
            if not os.path.isdir(self._path):
                os.makedirs(self._path)
            self.__load_items()
        finally:
            self._mutex.release()

    def length(self):
        self._mutex.acquire()
        try:
            return len(self._items)
        finally:
            self._mutex.release()

    def head(self):
        self._mutex.acquire()
        try:
            head_idx = self.__head_index
            if head_idx >= 0:
                logger.debug("head item: {}".format(head_idx))
                with open(self.__filename(head_idx), "r") as file:
                    return file.read()
            return None
        finally:
            self._mutex.release()

    def pop(self):
        self._mutex.acquire()
        try:
            text = None
            head_idx = self.__head_index
            if head_idx >= 0:
                logger.debug("pop item: {}".format(head_idx))
                with open(self.__filename(head_idx), "r") as file:
                    text = file.read()
                os.remove(self.__filename(head_idx))
                self._items.pop(0)
                # rotate items
                if self._max_head_index >= 0 and self.__head_index == self._max_head_index:
                    self.__rotate_items()
                # logger.debug("after pop items: {}".format(self._items))
            return text
        finally:
            self._mutex.release()

    def push(self, text):
        self._mutex.acquire()
        try:
            item = self.__tail_index + 1
            logger.debug("push item: {}".format(item))
            with open(self.__filename(item), "w") as file:
                file.write(text)
            self._items.append(item)
        except Exception:
            logger.error(traceback.format_exc())
            raise Exception("cannot push item")
        finally:
            self._mutex.release()

    @property
    def head_index(self):
        self._mutex.acquire()
        try:
            return self.__head_index
        finally:
            self._mutex.release()

    @property
    def max_head_index(self):
        return self._max_head_index

    @property
    def tail_index(self):
        self._mutex.acquire()
        try:
            return self.__tail_index
        finally:
            self._mutex.release()

    @property
    def __head_index(self):
        return self._items[0] if len(self._items) > 0 else -1

    @property
    def __tail_index(self):
        return self._items[len(self._items) - 1] if len(self._items) > 0 else -1

    def __load_items(self):
        self._items = [int(os.path.basename(file)[:-5]) for file in glob.glob(self._path + "/*.json")]
        self._items.sort()
        logger.debug("loaded items: {}".format(self._items))

    def __filename(self, item):
        return self._path + "/" + str(item) + ".json"

    def __rotate_items(self):
        logger.debug("rotating queue items: {}".format(self._items))
        counter = 0
        for item in self._items:
            # logger.debug("moving {} -> {}".format(self.__filename(item), self.__filename(counter)))
            os.rename(self.__filename(item), self.__filename(counter))
            counter = counter + 1
        self.__load_items()



# -*- encoding: utf-8 -*-
import sys
import os
import copy
import threading
import traceback
import logging
import time
import json

sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), '..', '..')))
import kamma
from kamma.queue import FileQueue


logger = logging.getLogger(__name__)


class KammaWorker(object):
    def __init__(self, tasks, queue_path="task_queue", interval_sec=15):
        self.interval_sec = max(1, interval_sec)
        self.quit = False
        self.thread = None
        self.mutex = threading.Lock()
        self.exception = None
        self.empty_event = threading.Event()
        # maps of task types --> task callback
        self.tasks = copy.copy(tasks)
        self.queue = FileQueue(queue_path)
        self.thread = threading.Thread(target=self._run, args=())
        self.thread.start()

    def __del__(self):
        self.stop()
        self.tasks = None
        self.queue = None

    def push_task(self, key, data):
        logger.debug('push task of type \'{}\' with data: {}'.format(key, data))
        if not self._is_registered(key):
            raise kamma.TaskNotRegistered('the task \'{}\' is not registered'.format(key))
        self.queue.push(json.dumps({'type': key, 'data': data}))
        self.empty_event.clear()

    def stop(self):
        logger.info('stopping')
        self.quit = True
        if self.thread and self.thread.is_alive():
            self.thread.join()
        logger.info('stopped')

    def wait(self, timeout=None):
        self.empty_event.wait(timeout)
        if self.exception:
            e = self.exception
            self.exception = None
            raise e

    def _is_registered(self, key):
        self.mutex.acquire()
        try:
            return True if key in self.tasks else False
        finally:
            self.mutex.release()

    def _get_callback(self, key):
        self.mutex.acquire()
        try:
            return self.tasks.get(key, None)
        finally:
            self.mutex.release()

    def _run(self):
        logger.info("running")
        try:
            while not self.quit:
                try:
                    self._process_queue()
                    if self.queue.length() == 0:
                        self.empty_event.set()
                    # wait cadence
                    for i in range(self.interval_sec):
                        if self.quit:
                            break
                        time.sleep(1)
                except kamma.TaskNotRegistered as e:
                    self.exception = e
                    self.empty_event.set()
                    raise self.exception
                except Exception:
                    logger.error(traceback.format_exc())
                    for i in range(self.interval_sec):
                        if self.quit:
                            break
                        time.sleep(1)
        finally:
            logger.info("exiting")

    def _process_queue(self):
        task_str = self.queue.head()
        count = 0
        while task_str and not self.quit:
            task = json.loads(task_str)
            key = task['type']
            callback = self._get_callback(key)
            if not callback:
                raise kamma.TaskNotRegistered('the task \'{}\' is not registered'.format(key))
            logger.debug('processing task: {}'.format(key))
            callback(task['data'])
            self.queue.pop()
            count = count + 1
            task_str = self.queue.head()
        if count > 0:
            logger.info("processed {} queued tasks".format(count))


if __name__ == "__main__":
    worker = KammaWorker()
    worker.stop()



# -*- encoding: utf-8 -*-

__version__ = '0.0.3'

import logging
try:
    # not available in python 2.6
    from logging import NullHandler
except ImportError:

    class NullHandler(logging.Handler):

        def emit(self, record):
            pass

# Add NullHandler to prevent logging warnings
logging.getLogger(__name__).addHandler(NullHandler())


class TaskNotRegistered(Exception):
    pass



from setuptools import setup
import kamma

long_description = ('kamma is a very simplified task file queue that '
                    'persist jobs and its input data and it has a worker that '
                    'process all pending jobs. Not recommended for high performance applications.')

setup(name='kamma',
      version=kamma.__version__,
      description='kamma is a very simplified Task File Queue',
      keywords=['kamma', 'file queue', 'task', 'job', 'task queue'],
      long_description=open('README.rst').read(),
      author='Marcel Janer Font',
      author_email='marceljanerfont@gmail.com',
      maintainer='Marcel Janer Font',
      maintainer_email='marceljanerfont@gmail.com',
      url='https://github.com/marceljanerfont/kamma',
      packages=['kamma'],
      license='MIT',
      package_data={'': ['LICENSE', 'README.rst']},
      classifiers=[
          'Development Status :: 4 - Beta',
          'Intended Audience :: Developers',
          'License :: OSI Approved :: MIT License',
          'Natural Language :: English',
          'Operating System :: OS Independent',
          'Programming Language :: Python :: 2.7',
          'Programming Language :: Python :: 3',
          'Programming Language :: Python :: 3.3',
          'Programming Language :: Python :: 3.4',
          'Programming Language :: Python :: 3.5',
          'Programming Language :: Python :: 3.6',
          'Programming Language :: Python :: Implementation :: CPython',
          'Programming Language :: Python :: Implementation :: Jython',
          'Programming Language :: Python :: Implementation :: PyPy',
          'Topic :: Communications',
          'Topic :: Internet',
          'Topic :: Software Development :: Libraries',
          'Topic :: Software Development :: Libraries :: Python Modules',
          'Topic :: System :: Networking'],
      zip_safe=True)
